import asyncio
from typing import Union

import discord
from cpl_core.configuration.configuration_abc import ConfigurationABC
from cpl_core.database.context.database_context_abc import DatabaseContextABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_query.extension import List
from discord.ext.commands import Context

from bot_core.abc.message_service_abc import MessageServiceABC
from bot_core.configuration.server_settings import ServerSettings
from bot_core.helper.log_message_helper import LogMessageHelper
from bot_core.logging.message_logger import MessageLogger
from bot_data.abc.client_repository_abc import ClientRepositoryABC


class MessageService(MessageServiceABC):

    def __init__(self, config: ConfigurationABC, logger: MessageLogger, bot: DiscordBotServiceABC, clients: ClientRepositoryABC, db: DatabaseContextABC):
        self._config = config
        self._logger = logger
        self._bot = bot
        self._clients = clients
        self._db = db
    
    async def delete_messages(self, messages: List[discord.Message], guild_id: int, without_tracking=False):
        self._logger.debug(__name__, f'Try to delete {messages.count()} messages')
        server_st: ServerSettings = self._config.get_configuration(f'ServerSettings_{guild_id}')
        await asyncio.sleep(server_st.message_delete_timer)
        for message in messages:
            await self.delete_message(message, mass_delete=True, without_tracking=without_tracking)
        self._logger.debug(__name__, 'Deleting messages finished')
        
    async def delete_message(self, message: discord.Message, mass_delete=False, without_tracking=False):
        server_st: ServerSettings = self._config.get_configuration(f'ServerSettings_{message.guild.id}')
        if not mass_delete:
            await asyncio.sleep(server_st.message_delete_timer)
        self._logger.debug(__name__, f'Try to delete message: {LogMessageHelper.get_log_string(message)}')
        guild_id = message.guild.id
        try:
            await message.delete()
            await asyncio.sleep(server_st.message_delete_timer)
        except Exception as e:
            self._logger.error(__name__, f'Deleting message failed', e)
        else:
            if not without_tracking:
                self._clients.append_deleted_message_count(self._bot.user.id, guild_id, 1)
                self._db.save_changes()
            self._logger.info(__name__, f'Deleted message {message}')
    
    async def send_channel_message(self, channel: discord.TextChannel, message: Union[str, discord.Embed], without_tracking=False):
        self._logger.debug(__name__, f'Try to send message\n\t{message}\n\tto: {channel}')
        msg = None
        try:
            if isinstance(message, discord.Embed):
                msg = await channel.send(embed=message)
            else:
                msg = await channel.send(message)
        except Exception as e:
            self._logger.error(__name__, f'Send message to channel {channel.id} failed', e)
        else:
            self._logger.info(__name__, f'Sent message to channel {channel.id}')
            if not without_tracking:
                self._clients.append_sent_message_count(self._bot.user.id, channel.guild.id, 1)
                self._db.save_changes()
            await self.delete_message(msg, without_tracking)
    
    async def send_dm_message(self, message: Union[str, discord.Embed], receiver: Union[discord.User, discord.Member], without_tracking=False):
        self._logger.debug(__name__, f'Try to send message\n\t{message}\n\tto: {receiver}')
        try:
            if isinstance(message, discord.Embed):
                msg = await receiver.send(embed=message)
            else:
                msg = await receiver.send(message)
        except Exception as e:
            self._logger.error(__name__, f'Send message to user {receiver.id} failed', e)
        else:
            if not without_tracking:
                self._clients.append_sent_message_count(self._bot.user.id, receiver.guild.id, 1)
                self._db.save_changes()
            self._logger.info(__name__, f'Sent message to user {receiver.id}')
    
    async def send_ctx_msg(self, ctx: Context, message: Union[str, discord.Embed], file: discord.File = None, is_persistent: bool = False, wait_before_delete: int = None, without_tracking=False):
        if ctx is None:
            self._logger.warn(__name__, 'Message context is empty')
            self._logger.debug(__name__, f'Message: {message}')
            return
        
        self._logger.debug(__name__, f'Try to send message\t\t{message}\n\tto: {ctx.channel}')
        msg = None
        try:
            if isinstance(message, discord.Embed):
                msg = await ctx.send(embed=message, file=file)
            else:
                msg = await ctx.send(message, file=file)
        except Exception as e:
            self._logger.error(__name__, f'Send message to channel {ctx.channel.id} failed', e)
        else:
            self._logger.info(__name__, f'Sent message to channel {ctx.channel.id}')
            if not without_tracking and ctx.guild is not None:
                self._clients.append_sent_message_count(self._bot.user.id, ctx.guild.id, 1)
                self._db.save_changes()

            if wait_before_delete is not None:
                await asyncio.sleep(wait_before_delete)

            if is_persistent:
                return

            if ctx.guild is not None:
                await self.delete_message(msg, without_tracking)
